/*
 * DeskLink for *nix (dl)
 * Copyright (C) 2004
 * Stephen Hurd
 *
 * Redistribution of modified and unmodified copies
 * is premitted provided the copyright remains intact
 */

#include <termios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#if defined(__APPLE__) && defined(__MACH__) && defined(__POWERPC__)
	#include <util.h>
#else
	#include <libutil.h>
#endif
#include <errno.h>

int file=-1;
int mode=0;	/* 0=unopened, 1=Write, 3=Read, 2=Append */
int term=-1;
unsigned char filename[25];
struct dirent *dire;
DIR *dir=NULL;
char **args;
struct termios origt;
struct termios	ti;
int gettymode;
int guimode=1;

unsigned char calc_sum(unsigned char type, unsigned char length, unsigned char *data)
{
	unsigned short sum=0;
	int i;

	sum+=type;
	sum+=length;
	for(i=0;i<length;i++)
		sum+=data[i];
	return((sum & 0xFF) ^ 255);
}

void normal_return(unsigned char type)
{
	unsigned char buf[4];

	buf[0]=0x12;
	buf[1]=0x01;
	buf[2]=type;
	buf[3]=calc_sum(0x12,0x01,buf+2);
	write(term,buf,4);
#ifdef DEBUG
	printf("Response: %02X\n",type);
#endif
}

int main(int argc, char **argv)
{
	int off=0;
	unsigned char *dir;
	unsigned char *p;
	unsigned char path[PATH_MAX];
	int arg;

	args=argv;
	dir=strdup(argv[0]);
	p=strrchr(dir,'/');
	if(p!=NULL) {
		*p=0;
		chdir(dir);
	}

	strcpy(path,"/dev/ttyd0");
	for(arg=1;arg<argc;arg++) {
		switch(argv[arg][0]) {
			case '/':
				strcpy(path,argv[1]);
				break;
			case '-':
				switch(argv[arg][1]) {
					case 0:
						strcpy(path,"/dev/tty");
						term=1;
						break;
					case 'g':
						gettymode=1;
						guimode=0;
						break;
					case 'c':
						guimode=0;
						break;
					default:
						printf("Unknown option %s\n",argv[arg]);
				}
				break;
			default:
				strcpy(path,"/dev/");
				strcat(path,argv[arg]);
		}
	}

	if(term<0)
		term=open(path,O_RDWR|O_NONBLOCK);
	if(term<0) {
		printf("Failed to open %s (%d)\n",path,errno);
		return(1);
	}

	if(gettymode) {
		if(login_tty(term)==0)
			term=STDIN_FILENO;
		else
			daemon(1,1);
	}

	(void)tcflush(term, TCIOFLUSH);	/* clear out the crap */
	ioctl(term, FIONBIO, &off);	/* turn off non-blocking mode */
	ioctl(term, FIOASYNC, &off);	/* ditto for async mode */

	if(tcgetattr(term,&ti)==-1) {
		printf("Error getting ttys settings %s %d (%d)\n",path,term,errno);
		return(1);
	}
	cfmakeraw(&ti);
	ti.c_cflag |= CLOCAL|CRTSCTS|CS8;
	if(cfsetspeed(&ti,B19200)==-1) {
		printf("Failed setting speed to %d on %s (%d)\n",B19200,path,errno);
		return(1);
	}
	if(tcsetattr(term,TCSANOW,&ti)==-1) {
		printf("Failed to set tty attributes %s %d (%d)\n",path,term,errno);
		return(1);
	}

	/* Set up terminal termios struct */
	origt.c_iflag=BRKINT | ICRNL | IMAXBEL | IXON | IXANY | IXOFF;
	origt.c_oflag=OPOST | ONLCR;
	origt.c_lflag=ECHO | ICANON | ISIG | IEXTEN | ECHOE | ECHOKE | ECHOCTL;
	origt.c_cflag=CREAD | CS8 | HUPCL | CCTS_OFLOW | CRTS_IFLOW;
	cfsetspeed(&origt,B1200);

	while(1)
		bedisk();

	return(0);
}

int send_dirent (unsigned char *buf, struct stat *st)
{
	unsigned short size;
	unsigned char *dot;
	unsigned char *p;

	if(dire!=NULL) {
		buf[26]='F';
		size=htons(st->st_size);
		memcpy(buf+27,&size,2);
		memset(buf+2,' ',24);
		memcpy(buf+2,dire->d_name,strlen(dire->d_name));
		dot=memchr(buf+2,'.',24);
		if(dot!=NULL) {
			memcpy(dot,buf+8,3);
			for(p=dot;p<buf+8;p++)
				*p=' ';
		}
	}
	buf[30]=calc_sum(0x11,0x1C,buf+2);
	write(term,buf,31);
}

int read_next_dirent(struct stat *st)
{
	char *dot;

	if(dir==NULL) {
		dire=NULL;
		normal_return(0x70);
		return(0);
	}
	while((dire=readdir(dir))!=NULL) {
		if(dire->d_type!=DT_REG)
			continue;

		if(stat(dire->d_name,st)) {
			normal_return(0x31);
			return 0;
		}

		if(st->st_size>65535)
			continue;

		/* Check if "legal" name XXXXXX.YY */
		dot=strchr(dire->d_name,'.');

		if(dot==NULL && strlen(dire->d_name)>6)
			/* Not dot, name more than 6 chars */
			continue;
		else {
			/* More than 2 chars after the dot */
			if(dot!=NULL && strlen(dot)>3)
				continue;

			/* More than 6 chars before the dot */
			if(dot!=NULL && dot>dire->d_name+6)
				continue;
		}

		break;
	}
	return 1;
}

int directory(unsigned char length, unsigned char *data)
{
	unsigned char search_form;
	int i;
	unsigned char *p,*dot;
	unsigned char buf[31];
	struct stat st;

	bzero(buf,31);
	buf[29]=40;
	buf[0]=0x11;
	buf[1]=0x1C;
	search_form=data[length-1];
	switch (search_form) {
		case 0x00:	/* Pick file for open/delete */
			strncpy(filename,data,24);
			filename[25]=0;
			/* Remove trailing spaces */
			for(p=strrchr(filename,' ');p>=filename && *p==' ';p--)
				*p=0;
			/* Remove spaces between base and dot */
			dot=strchr(filename,'.');
			if(dot!=NULL) {
				for(p=dot-1;*p==' ' && p>filename;p--)
					bcopy(dot--,p,strlen(dot)+1);
			}
			if(dir!=NULL)
				closedir(dir);
			dir=opendir(".");
			for(;read_next_dirent(&st) && dire!=NULL && strcmp(dire->d_name,filename););
				send_dirent(buf,&st);
			break;
		case 0x01:	/* "first" directory block */
			if(dir!=NULL)
				closedir(dir);
			dir=opendir(".");
			if(read_next_dirent(&st))
				send_dirent(buf, &st);
			break;
		case 0x02:	/* "next" directory block */
			if(read_next_dirent(&st))
				send_dirent(buf, &st);
			break;
		case 0x03:	/* "previous" directory block */
			break;
		case 0x04:	/* end directory reference */
			closedir(dir);
			break;
	}
	return 0;
}

int open_file(unsigned char omode)
{
	struct stat st;

#ifdef DEBUG
	printf("Open Mode: %d\n",omode);
#endif
	switch(omode) {
		case 0x01:	/* New file for write */
			if(file>=0) {
				close(file);
				file=-1;
			}
			file=open(filename,O_CREAT|O_TRUNC|O_WRONLY|O_EXCL,0666);
			if(file<0)
				normal_return(0x37);
			else {
				mode=omode;
				normal_return(0x00);
			}
			break;
		case 0x02:	/* existing file for append */
			if(file>=0) {
				close(file);
				file=-1;
			}
			stat(filename,&st);
			if(st.st_size>65535) {
				normal_return(0x6E);
				break;
			}
			file=open(filename,O_WRONLY|O_APPEND);
			if(file<0)
				normal_return(0x37);
			else {
				mode=omode;
				normal_return(0x00);
			}
			break;
		case 0x03:	/* Existing file for read */
			if(file>=0) {
				close(file);
				file=-1;
			}
			stat(filename,&st);
			if(st.st_size>65535) {
				normal_return(0x6E);
				break;
			}
			file=open(filename,O_RDONLY);
			if(file<0)
				normal_return(0x37);
			else {
				mode=omode;
				normal_return(0x00);
			}
			break;
	}
}

void read_file(void)
{
	unsigned char buf[131];
	int in;

	buf[0]=0x10;
	if(file<0) {
		normal_return(0x30);
		return;
	}
	if(mode!=3) {
		normal_return(0x37);
		return;
	}
	in=read(file,buf+2,128);
	buf[1]=(unsigned char)in;
	buf[2+in]=calc_sum(0x10, (unsigned char)in, buf+2);
	write(term,buf,3+in);
}

void renamefile(unsigned char *name)
{
	unsigned char *p;

	name[24]=0;
	for(p=strrchr(name,' ');*p=' ';p--)
		*p=0;
	if(rename(filename,name))
		normal_return(0x4A);
	else
		normal_return(0x00);
}

int readbytes(int handle, void *buf, int max)
{
	int r=0;
	int i=0;

	while(r<max)
		r+=read(term,buf+r,1);

	return(r);
}

int sendloader(void)
{
	int w=0;
	int i=0;
	int fd;
	unsigned char b;

	if((fd=open("loader.ba",O_RDONLY))<0) {
#ifdef DEBUG
		printf("Failed to open loader.ba for read.\n");
#endif
		return(9);
	}

	sleep(1);

	while(read(fd,&b,1)==1) {
		while((i=write(term,&b,1))!=1);
		w+=i;
		usleep(5000);
#ifdef DEBUG
		printf("\rSent: %d bytes",w);
		fflush(stdout);
#endif
	}
	b=0x1A;
	write(term,&b,1);
	close(fd);
	close(term);
#ifdef DEBUG
	puts("");
	if(strlen(args[0])>4)
		strcpy(args[0],"./dl");
	execvp("./dl",args);
#else
	exit(0);
#endif

	return(0);
}

int bedisk(void)
{
	unsigned char preamble[3];
	unsigned char type;
	unsigned char length;
	unsigned char data[0x81];
	unsigned char checksum;
	int  r=0;

	preamble[2]=0;

	readbytes(term, preamble, 2);

	if(preamble[0]==preamble[1]) {
		switch(preamble[0]) {
			case 'Z':
				break;
			case 'Y':
				tcsetattr(term,TCSANOW,&origt);
				system("login");
				tcsetattr(term,TCSANOW,&ti);
				return(10);
			case 'X':
				sendloader();
				return(sendloader());
			default:
#ifdef DEBUG
				printf("Bad preamble: '%s' (%02X %02X)\n",preamble,preamble[0],preamble[1]);
#endif
				normal_return(0x40);
				return(2);
		}
	}
	else {
#ifdef DEBUG
		printf("Bad preamble: '%s' (%02X %02X)\n",preamble,preamble[0],preamble[1]);
#endif
	}


	readbytes(term,&type,1);
	
	readbytes(term,&length,1);
	
	readbytes(term,data,length);
	data[length]=0;
	
	readbytes(term,&checksum,1);

	if(calc_sum(type,length,data)!=checksum) {
#ifdef DEBUG
		printf("BAD CHECKSUM! Preamble: %s Type: %02X Length: %02X Data: %s\n",preamble,type,length,data);
		printf("Packet checksum: %02X  My checksum: %02X\n",checksum,calc_sum(type,length,data));
#endif
		normal_return(0x36);
		return(7);
	}

#ifdef DEBUG
	printf("Preamble: %s Type: %02X Length: %02X Data: %s\n",preamble,type,length,data);
#endif
	switch(type) {
		case 0x00:	/* Directory ref */
			directory(length,data);
			break;
		case 0x01:	/* Open file */
			open_file(data[0]);
			break;
		case 0x02:	/* Close file */
			if(file>=0)
				close(file);
			normal_return(0x00);
			break;
		case 0x03:	/* Read */
			read_file();
			break;
		case 0x04:	/* Write */
			if(file<0) {
				normal_return(0x30);
				break;
			}
			if(mode!=1 && mode !=2) {
				normal_return(0x37);
				break;
			}
			if(write(file,data,length)!=length)
				normal_return(0x4a);
			else
				normal_return(0x00);
			break;
		case 0x05:	/* Delete */
			unlink(filename);
			normal_return(0x00);
			break;
		case 0x06:	/* Format disk */
			normal_return(0x00);
			break;
		case 0x07:	/* Drive Status */
			normal_return(0x00);
			break;
		case 0x08:	/* Drive condition */
			normal_return(0x00);
			break;
		case 0x0D:	/* Rename File */
			renamefile(data);
			break;
		default:
			return(8);
			break;
	}
	return(0);
}
